/***************************************************************************
 *   Copyright (C) 2004 by Spiros Georgaras                                *
 *   sngeorgaras@otenet.gr                                                 *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser Public License as published by   *
 *   the Free Software Foundation; either version 2.1 of the License, or   *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser Public License for more details.                           *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser Public License      *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/stat.h> 
#include <unistd.h> 
#include <grp.h>
#include <pwd.h>
#include <stdlib.h>

#include <iostream>
#include <kmimetype.h>
#include <ksharedptr.h>
#include <kmdcodec.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <ktempdir.h> 
#include <kstandarddirs.h>
#include <krun.h>
#include <qdatastream.h>
#include <qtextstream.h> 
#include <qdir.h>
#include <qregexp.h>
#include <kdebug.h>

#include <kmhtmlarchive.h>
#include <qglobal.h>
/*================================
================================*/
/*KmhtmlArchive::KmhtmlArchive(){
	item=0;
	status=0;
	isDOS=FALSE;
	parts=0;
	controlChars[0].sprintf("%c",13);
	controlChars[1].sprintf("%c",10);
	controlChars[2].sprintf("%c",9);
}
*/
/*================================
================================*/
KmhtmlArchive::KmhtmlArchive(QString mhtArchive){
	item=0;
	status=0;
	mhtFileName=mhtArchive;
	isDOS=FALSE;
	parts=-1;
	controlChars[0].sprintf("%c",13);
	controlChars[1].sprintf("%c",10);
}
/*================================
================================*/
KmhtmlArchive::~ KmhtmlArchive(){
	status=0;
	items=0;
	for(int i=0;i<items;i++) delete [] item[i].data;
	delete [] item;
	if(parts>=0) delete [] partHeader;
}
 /*================================
================================*/
void KmhtmlArchive::reset(){
	//kdDebug()<<"Reseting"<<endl;
	status=0;
	items=0;
	for(int i=0;i<items;i++) delete [] item[i].data;
	delete [] item;
	if(parts>=0) delete [] partHeader;
	allocItems=-1;
	parts=-1;
	f="";
}
/*================================
================================*/
bool KmhtmlArchive::open(int mode){
	ulong filesize;
	ulong flsize;
	if(mode==IO_ReadOnly){
		//kdDebug() << "[KmhtmlArchive::open()]: mhtFileName = " << mhtFileName << endl;
		file.setName(mhtFileName);
		if(!file.open( mode )){
			emit errorAppeared(1);
			status=1;
			//kdDebug() << "[KmhtmlArchive::open()]: Error opening file" << endl;
			return FALSE;
		}else emit archiveOpened();
		m_size=(long) file.size();
		if(!m_size){
			emit errorAppeared(3);
			status=3;
			return FALSE;
		}
		filesize=(ulong)file.size();
		flsize=filesize/10;
		if(!flsize){
			status=3;
			emit errorAppeared(3);
			return FALSE;
		}
		f="";
		QDataStream stream( &file );  // read the data serialized from the file
		while (!stream.atEnd()) stream>>f;
		if(!readPartHeader()){
			//kdDebug() << "[KmhtmlArchive::open()]: Error reading header" << endl;
			emit errorAppeared(2);
			status=2;
			return FALSE;
		}
		if(!readItems()){
			status=3;
			emit errorAppeared(3);
			//kdDebug() << "[KmhtmlArchive::open()]: Error reading items" << endl;
			return FALSE;
		}
		//kdDebug() << "[KmhtmlArchive::open()]: done" << endl;
		//return FALSE; //test case
		// read time,user,group,access
		struct stat statbuf;
		::stat( QFile::encodeName( mhtFileName ), &statbuf );
		m_mtime = statbuf.st_mtime;
		struct passwd* pwp;
		pwp=getpwuid(statbuf.st_uid);
		m_user=QString(pwp->pw_name);
		struct group* grp;
		grp = getgrgid(statbuf.st_gid);
		m_group=grp->gr_name;
		//m_permissions=statbuf.st_mode & 07777;
		m_permissions=statbuf.st_mode;
		return TRUE;
	}else if(mode==IO_WriteOnly){
		return FALSE;
	}else return FALSE;
}
/*================================
================================*/
void KmhtmlArchive::setFileName(QString fileName){
	mhtFileName=fileName;
}
/*================================
================================*/
bool KmhtmlArchive::readItems(){
	int i=0;
	items=0;
	int it; // will have the final number of items
	int newItems;
	delete [] item;
	item=new elm[5];
	for(int o=0;o<5;o++){
		item[o].data=0;
		item[o].fixed=FALSE;
		item[o].level=0;
	}
	allocItems=5;
	if(!readItemsHeader(&f,0)){
		//kdDebug()<<"[KmhtmlArchive::readItems()]: error in readItemsHeader"<<endl;
		return FALSE;
	}
	//kdDebug()<< "[KmhtmlArchive::readItems()]: items="<<items<<endl;
	// search for nested parts
	it=items;
	for(i=0;i<items;i++){
		//break; // test case
		if(item[i].header.find("multipart/")!=-1){
			//kdDebug()<< " ---------- [KmhtmlArchive::readItems()]: found nested block in item " << i << endl;
			item[i].isInvalid=TRUE;
			it--;
		}else{
			item[i].isInvalid=FALSE;
			//kdDebug()<< "[KmhtmlArchive::readItems()]: item["<<i<<"] header= "<<item[i].header<<endl;
		}
	}
	if(it!=items){ // if nested parts found add them to the partsHeader
		// reallocate partHeader
		struct HEADER* tmp;
		tmp=new HEADER[parts+items-it];
		for(i=0;i<parts;i++){
			tmp->start=partHeader->start;
			tmp->end=partHeader->end;
			tmp->itStart=partHeader->itStart;
			tmp->header=partHeader->header;
			tmp->contenttype=partHeader->contenttype;
			tmp->contentLocation=partHeader->contentLocation;
			tmp->type=partHeader->type;
			tmp->boundary=partHeader->boundary;
		}
		delete [] partHeader;
		partHeader=tmp;
		int l=0;
		for(i=0;i<items;i++){
			if(item[i].isInvalid){
				partHeader[parts+l].header=item[i].header;
				partHeader[parts+l].start=item[i].start;
				partHeader[parts+l].end=item[i].end;
				partHeader[parts+l].itStart=item[i].itStart;
				//kdDebug()<< "[KmhtmlArchive::readItems()]: parts=" << parts << endl;
				l++;
			}
		}
		//kdDebug()<< "[KmhtmlArchive::readItems()]: items=" << items << endl;
		//kdDebug()<< "[KmhtmlArchive::readItems()]: it=" << it << endl;
		newItems=items-it;
		//kdDebug()<< "[KmhtmlArchive::readItems()]: newItems=" << newItems << endl;
		parts=parts+newItems;
		//kdDebug()<< "[KmhtmlArchive::readItems()]: parts=" << parts << endl;

// 		for(i=0;i<parts;i++){
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]: start=" << partHeader[i].start << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]: itStart=" << partHeader[i].itStart << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]: end=" << partHeader[i].end << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]: part " << i << " header" << endl;
// 			kdDebug()<< partHeader[i].header << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 			kdDebug()<< "[KmhtmlArchive::readItems()]" << endl;
// 		}

		// remove them from items
		struct elm* tmp1;
		if(it>0){
			tmp1=new elm[it];
			allocItems=it;
		}else{
			tmp1=new elm[5];
			allocItems=5;
		}
		int o=0;
		for(i=0;i<items;i++){
			if(!item[i].isInvalid){
				tmp1[o].start=item[i].start;
				tmp1[o].end=item[i].end;
				tmp1[o].itStart=item[i].itStart;
				tmp1[o].header=item[i].header;
				tmp1[o].level=item[i].level;
				tmp1[o].isInvalid=FALSE;
 				//kdDebug()<< "[KmhtmlArchive::readItems()]: copying to item " << o << ": " << item[i].header<<endl;
				o++;
			}
		}
		delete [] item;
		item=tmp1;
		items=it;
 		//kdDebug()<< "[KmhtmlArchive::readItems()]: items=" << items << endl;
		// read new part headers
		for(i=parts-newItems;i<parts;i++){
			//kdDebug()<<"i="<<i<<", parts="<<parts<<endl;
			if(!readPartHeader(i)){
				//kdDebug() << "[KmhtmlArchive::open()]: Error reading header" << endl;
				status=2;
				return FALSE;
			}
		}
		// find new items in the nested parts	
		//kdDebug() << "[KmhtmlArchive::open()]: parts=" << parts << " newItems=" << newItems << " diff=" << parts-newItems << endl;
		for(i=parts-newItems;i<parts;i++){
			//kdDebug()<< "[KmhtmlArchive::readItems()]: counting = " << i << endl;
			//kdDebug()<< "[KmhtmlArchive::readItems()]: items before reading= " << items << endl;
			QCString inner=f.mid(partHeader[i].start,partHeader[i].end-partHeader[i].start);
			if(!readItemsHeader(&inner,i)) return FALSE;
			//kdDebug()<< "[KmhtmlArchive::readItems()]: items after reading= " << items << endl;
		}
	}
	//kdDebug()<< "[KmhtmlArchive::readItems()]: total items = " << items << endl;
	for(i=0;i<items;i++){
		if(item[i].level>0){
			//kdDebug()<<"[KmhtmlArchive::readItems()]: level=" << item[i].level << " i="<< i << " item[i].level" <<item[i].level << " partHeader[item[i].level].start=" << partHeader[item[i].level].start << endl;
			item[i].start+=partHeader[item[i].level].start;
			item[i].itStart+=partHeader[item[i].level].start;
			item[i].end+=partHeader[item[i].level].start;
		}
		//kdDebug() << "[KmhtmlArchive::readItems()] -----------------" << endl << "item[" << i << "] with start "<< item[i].start << ", itStart "<< item[i].itStart <<", end "<< item[i].end<< " and level=" << item[i].level << endl << item[i].header << endl;
		//kdDebug() << "-----------------" << endl;
		evalItemHeader(i);
	}

	for(int i=0;i<items;i++){
		QCString fContents;
		fContents=f.mid(item[i].itStart,item[i].end-item[i].itStart+1);
		QByteArray exC(100);
		QCString exContents=fContents;
		QByteArray inC(100);
		switch(item[i].encoding){
		case 0: // quoted printable
			exContents=removeControlChars(KCodecs::quotedPrintableDecode(fContents));
			fixFileName(fContents,i);
			item[i].data=exContents;
			item[i].decodedSize=exContents.length();
			break;;
		case 1: // base64
			inC.setRawData( fContents.data(), fContents.length() );
 			KCodecs::base64Decode( inC, exC );
			inC.resetRawData( fContents.data(), fContents.length() );
			fixFileName(exC,i);
			item[i].data=exC;
			item[i].decodedSize=exC.size();
			//kdDebug()<<"item[i].decodedSize="<<item[i].decodedSize<<endl;
			break;;
		case 2: // uuencode
			inC.setRawData( fContents.data(), fContents.length() );
 			KCodecs::uudecode( inC, exC );
			inC.resetRawData( fContents.data(), fContents.length() );
			fixFileName(exC,i);
			item[i].data=exC;
			item[i].decodedSize=exC.size();
			break;;
		case 3: // 7-bit
		case 4: // 8-bit
			item[i].data=removeControlChars(fContents);
			item[i].decodedSize=fContents.length();
			break;;
		}
	}
	// get main doc filename
	for(int i=0;i<items;i++){
		//if(item[i].mimeType==type){
			if(item[i].mimeType=="text/html") item[i].alias="index.html";
			else if (item[i].mimeType=="text/xml") item[i].alias="index.xml";
			else if (item[i].mimeType=="application/xhtml+xml") item[i].alias="index.xhtml";
		//}
	}
	// create alias
	for(int i=0;i<items;i++){
		if(item[i].alias.isEmpty()) item[i].alias=item[i].fixedName;
		//kdDebug()<< "[KmhtmlArchive::readData]: alias for "<< i << " " << item[i].name << "," << item[i].fixedName << "," << item[i].alias << endl;
	}
	// fix dublicate aliases
	for(i=0;i<items;i++){
		int cr=1;
		for(int y=i+1;y<items;y++){
			if(item[i].alias==item[y].alias){
				int l=item[y].alias.find(".");
				QString g;
				g.setNum(cr);
				g.prepend(item[y].alias.left(l));
				g.append(item[y].alias.mid(l));
				item[y].alias=g;
				cr++;
			}
		}
	}
	
	// find type
	type="";
	for(int i=0;i<items;i++){
		if(item[i].mimeType=="text/html" || item[i].mimeType=="text/xml" || item[i].mimeType=="application/xhtml+xml"){
			type=item[i].mimeType;
			//kdDebug() << "type=" << type << endl;
			break;
		}
	}
	if(type.isEmpty()) type=QString("inode/directory");
	
	// check for mixed part
// 	for(int y=0;y<parts;y++){
// 		if(partHeader[y].contenttype=="multipart/mixed"){
// 			kdDebug()<< "[KmhtmlArchive::readData]: multipart/mixed found at " << y << endl;
// 			// if multipart/mixed, combine html files....
// 			int markItems=-1;
// 			int ind[5];
// 			int id=0;
// 			for(int i=0;i<items;i++){
// 				kdDebug()<<"i="<<i<<" level="<<item[i].level<<endl;
// 				if(item[i].level>y && item[i].mimeType==type){
// 					ind[id]=i;
// 					id++;
// 					markItems++;
// 				}
// 			}
// 			kdDebug()<< "[KmhtmlArchive::readData]: markItems="<<markItems<<endl;
// 			if(markItems<1) continue;
// 			QCString out;
// 			//first item
// 			QCString tmp=f.mid(item[ind[0]].itStart,item[ind[0]].end-item[ind[0]].itStart+1);
// 			//kdDebug()<< "[KmhtmlArchive::readData]: tmp="<<tmp<<endl;
// 			int l=tmp.find("</html>",0,FALSE);
// 			//kdDebug()<< "[KmhtmlArchive::readData]: l="<<l<<endl;
// 			if(l!=-1) out=f.mid(item[ind[0]].itStart,l-item[ind[0]].itStart);
// 			else out=tmp;
// 			
// 			// last item
// 			out+="<br><hr><br>";
// 			kdDebug()<< "[KmhtmlArchive::readData]: out="<<out<<endl;
// 			tmp=f.mid(item[ind[id]].itStart,item[ind[id]].end-item[ind[id]].itStart+1);
// 			kdDebug()<< "[KmhtmlArchive::readData]: tmp="<<tmp<<endl;
// 			l=tmp.find("<body",0,TRUE);
// 			if(l!=-1) out+=f.mid(l,item[ind[0]].end-l);
// 			else out+=tmp;
// 			
// 			kdDebug()<<"out"<<endl<<out<<endl;
// 			
// 			//break;
// 		}
// 	}
// 	
	emit foundFilesInArchive(items);
	//kdDebug()<< "[KmhtmlArchive::readData]: items=" << items << endl;
	m_entries.clear();
	for(i=0;i<items;i++) m_entries.insert(m_entries.end(),item[i].alias);
	//return FALSE; // test case
	return TRUE;
}
/*================================
================================*/
bool KmhtmlArchive::readItemsHeader(QCString* cont,int count){
	int i=0; 
	int j=0;
	int prevItems=items;
	QCString b=partHeader[count].boundary;
	if(b.isEmpty()){
		//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: b.isEmpty(), count =" << count << endl;
		items=1;
		item[0].start=0;
		item[0].header=partHeader[count].header;
		i=item[0].header.length();
		while(cont->mid(i,1)==controlChars[0] || cont->mid(i,1)==controlChars[1]) i++;
		item[0].itStart=i;
		//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: item[0].itStart = " << item[0].itStart << endl;
		item[0].end=cont->length();
	}else{
		while((j=cont->find(b.data(),j))!=-1){
			//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: j="<<j<<endl;
			if((uint)j>cont->length()) break;
			items++;
			j+=b.length();
			i=j;
		}
		uint lastBoundary=i;
		items-=2; // don't count the header and the last instance
		//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: items=" << items << " count="<<count<< endl;
		if(!items) return FALSE;
		//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: items=" << items << endl;
		// reallocate items
		if(items>allocItems){
			// reallocate item
			struct elm* tpmItem;
			tpmItem=new elm[items];
			for(int o=0;o<prevItems;o++){
				tpmItem[o].start=item[o].start;
				tpmItem[o].end=item[o].end;
				tpmItem[o].itStart=item[o].itStart;
				tpmItem[o].header=item[o].header;
				tpmItem[o].level=item[o].level;
				tpmItem[o].isInvalid=FALSE;
				// addition
				tpmItem[o].encodedSize=item[o].encodedSize;
				tpmItem[o].decodedSize=item[o].decodedSize;
				tpmItem[o].contentLocation=item[o].contentLocation;
				tpmItem[o].contentID=item[o].contentID;
				tpmItem[o].dispFileName=item[o].dispFileName;
				tpmItem[o].encoding=item[o].encoding;
				tpmItem[o].charset=item[o].charset;
				tpmItem[o].mimeType=item[o].mimeType;
				tpmItem[o].hasExtension=item[o].hasExtension;
				tpmItem[o].fixed=item[o].fixed;
				tpmItem[o].name=item[o].name;
				tpmItem[o].fixedName=item[o].fixedName;
				tpmItem[o].alias=item[o].alias;
				tpmItem[o].data=item[o].data;
			}
			for(int o=0;o<items;o++){
				tpmItem[o].data=0;
				tpmItem[o].fixed=FALSE;
				//tpmItem[o].level=0;
			}
			delete [] item;
			item=tpmItem;
			allocItems=items;
		}
		
		//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: i+length = " << i << ", file length = " << cont->length() << endl;
		//kdDebug()<< "-------------- [KmhtmlArchive::readItemsHeader()]: items = " << items << endl;
		
	/*	
		//check end of file
		// After the last found boundary I can accept --
		QCString a;
		a.sprintf("%c",0x2d);
		if(cont->mid(i,1)!=a || cont->mid(i+1,1)!=a){
			//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: " << cont->mid(i,1) << ", " << cont->mid(i+1,1)<< endl;
			emit errorAppeared(2); //FIX ?
			status=2;
			//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: invalid end of file" << endl;
			return FALSE;
		}
		*/
		
		i=cont->find(b);
		// find item start
		for(j=prevItems;j<items;j++){
			i=cont->find(b,i+1);
			item[j].start=i+b.length();
			if(isDOS) item[j].start+=2;
			else item[j].start++;
			if(j>prevItems && j<items){
				item[j-1].end=item[j].start-b.length();
				//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: item "  << j-1 << " end = " << (uint)item[j-1].end << endl;
				if(cont->mid(item[j-1].end,1)!=controlChars[0] && cont->mid(item[j-1].end,1)!=controlChars[1]){
					item[j-1].end--;
					while(cont->mid(item[j-1].end,1)!=controlChars[0] && cont->mid(item[j-1].end,1)!=controlChars[1]){
						item[j-1].end--;
						//kdDebug()<<" ! item[j-1].end="<<(uint)item[j-1].end<<endl;
					}
				}
				while(cont->mid(item[j-1].end,1)==controlChars[0] || cont->mid(item[j-1].end,1)==controlChars[1]){
					item[j-1].end--;
					//kdDebug()<<"   item[j-1].end="<<(uint)item[j-1].end<<endl;
				}
				//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: item "  << j-1 << " end = " << (uint)item[j-1].end << endl;
			}
			//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: item "  << j << " start = " << (uint)item[j].start << endl;
		}
		item[items-1].end=lastBoundary-b.length();
		item[items-1].end=cont->findRev(10,item[items-1].end);
		while(cont->mid(item[items-1].end,1)==controlChars[0] || cont->mid(item[items-1].end,1)==controlChars[1]) item[items-1].end--;
		//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: item "  << items-1 << " end = " << (uint)item[items-1].end << endl;
		//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: item "  << j-1 << " end = " << (uint)item[j-1].end << endl;
		
		for(i=prevItems;i<items;i++){
			item[i].level=count;
			//kdDebug()<<"[KmhtmlArchive::readItemsHeader]: item[" << i << "] level = " << count << endl;
			j=item[i].start;
			if(isDOS){
				bool found=FALSE;
				while(!found){
					//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: before j="  << j << endl;
					j=cont->find(10,j+1);
					//if(j>item[i].end) found=TRUE;
					//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: after j="  << j << endl;
					if(cont->mid(j+1,1)==controlChars[0] && cont->mid(j+2,1)==controlChars[1]){
						item[i].header=removeControlChars(cont->mid(item[i].start,j+2-item[i].start));
						found=TRUE;
						item[i].itStart=j+2;
						//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: itStart ="  << item[i].itStart << endl;
					}
				}
			}else{
				//kdDebug()<< "[KmhtmlArchive::readItemsHeader()]: ! isDOS" << endl;
				bool found=FALSE;
				while(!found){
					//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: before j="  << j << endl;
					j=cont->find(10,j+1);
					//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: after j="  << j << endl;
					if(cont->mid(j+1,1)==controlChars[1]){
						//kdDebug() << "[KmhtmlArchive::readItemsHeader()] found!!!" << endl;
						//kdDebug() << "item[i].start=" << item[i].start << " j=" << j << endl;
						//kdDebug() <<removeControlChars(cont->mid(item[i].start,j+2-item[i].start))<< endl;
						item[i].header=removeControlChars(cont->mid(item[i].start,j+2-item[i].start));
						//kdDebug() << "[KmhtmlArchive::readItemsHeader()] header set!!!" << endl;
						found=TRUE;
						item[i].itStart=j+2;
						//kdDebug() << "[KmhtmlArchive::readItemsHeader()]: itStart ="  << item[i].itStart << endl;
					}
				}
			}
		}
	}
	// return FALSE; // test case
	return TRUE;
}
/*================================
================================*/
QDateTime KmhtmlArchive::datetime() const{
	QDateTime d;
	d.setTime_t( m_mtime );
	return d;
}
/*================================
================================*/
void KmhtmlArchive::fixFileName(QByteArray exC,int i){
	QString ext=QString::null;
	QString tmp=item[i].name;
	if(tmp.isEmpty()){
		// if no name candidate found, fall back to mimetype
		int c=item[i].mimeType.find("/");
		tmp=item[i].mimeType.left(c);
		//kdDebug() << "[KmhtmlArchive::fixFileName]: tmp=" << tmp << endl;
	}
	if(item[i].mimeType=="application/octet-stream"){
		int accuracy=0;
		KMimeType::Ptr mt = KMimeType::findByContent(exC, &accuracy );
		if(mt->name().left(4)!="text") item[i].mimeType=mt->name();
	}
	KMimeType::Ptr ptr;
	ptr= KMimeType::mimeType(item[i].mimeType );
	//kdDebug() << "[KmhtmlArchive::fixFileName]: ptr=" << ptr->name() << endl;
	if(ptr->name()=="application/octet-stream"){
		int accuracy=0;
		ptr = KMimeType::findByContent(exC, &accuracy );
		//kdDebug() << "[KmhtmlArchive::fixFileName]: fixxed ptr=" << ptr->name() << endl;
	}
	QStringList patterns=ptr->patterns();
	for ( QStringList::Iterator it = patterns.begin(); it != patterns.end(); ++it ) {
		if(ext.isNull()){
			ext=*it;
			ext=ext.mid(2);
		}
		//kdDebug() << "[KmhtmlArchive::fixFileName]: ext=" << ext << endl;
		if(!ext.isNull() && ext.length()>3){
			QString text=(*it).mid(2);
			//kdDebug() << "[KmhtmlArchive::fixFileName]: text=" << text << endl;
			if(text.length()<4){
				ext=text;
				break;
			}
		}
	}
	ext=ext.lower();
	//kdDebug() << "[KmhtmlArchive::fixFileName]: extension=" << ext << " item=" << i << " name="  << tmp << endl;
	if(!item[i].hasExtension){
		tmp=tmp.append("." +ext);
		item[i].hasExtension=TRUE;
	}else{
		if(!ext.isEmpty() && tmp.right(4).lower()!=QString(".css") && tmp.right(3).lower()!=QString(".js")){
			int c=tmp.findRev(".");
			tmp=tmp.left(c+1)+ext;
		}
	}
	ptr= KMimeType::mimeType(item[i].mimeType );
	item[i].fixedName=tmp;
	//kdDebug() << "[KmhtmlArchive::fixFileName]: alias=" << item[i].alias << endl;
}
/*================================
================================*/
bool KmhtmlArchive::readPartHeader(int it){
	bool errorFound=FALSE;
	int i=0;
	if(it==-1){
		parts=0;
		partHeader=new HEADER[1];
		isDOS=findLineEnd();
		int l=-1;
		int c=0;
		bool found=FALSE;
		while(!found){
			i=f.find(10,i+1);
			if(c==0) l=i;
			else{
				if(l==i){
					emit errorAppeared(3);
					status=3;
					return FALSE;
				}
			}
			c++;
			if(isDOS && f.mid(i+1,1)==controlChars[0] && f.mid(i+1,1)==controlChars[0]) found=TRUE;
			else if(!isDOS && f.mid(i+1,1)==controlChars[1]) found=TRUE;
		}
		while(f.mid(i,1)==controlChars[0] || f.mid(i,1)==controlChars[1]) i--;
		partHeader->header=f.mid(0,i+1);
		parts++;
		//kdDebug() << "[KmhtmlArchive::readPartHeader]: parts=" << parts << endl;
		it=parts-1;
	}
	//kdDebug(  ) << "[KmhtmlEntry::readPartHeader]" << endl << partHeader[it].header << endl;
	if(errorFound){
			//kdDebug(  ) << "[KmhtmlEntry::readHeader]: errorFound" << endl;
			emit errorAppeared(2);
			status=2;
			return FALSE;
	}
	//read boundary
	QCString msg="boundary=";
	i=partHeader[it].header.find(msg);
	if(i==-1){
		partHeader[it].boundary="";
	}else{
		i=partHeader[it].header.find('"',i);i++;
		int j=partHeader[it].header.find('"',i);
		partHeader[it].boundary=partHeader[it].header.mid(i,j-i);
		if(it==0){
			int h=f.length()-1;
			while(f.mid(h,1)==controlChars[0] || f.mid(h,1)==controlChars[1]) h--;
			if(!f.mid(h,1)==controlChars[0] || !f.mid(h,1)==controlChars[1]){
				emit errorAppeared(3);
				status=3;
				return FALSE;
			}
		}
	}
	//kdDebug()<<"[KmhtmlEntry::readPartHeader]: part="<<it<<" header="<<partHeader[it].header<<endl;
	msg="Content-Type:"; // --------------- read Content-Type
	i=partHeader[it].header.find(msg,0,TRUE);
	if(i==-1) partHeader[it].contenttype="";
	else{
		i++; // to include space ": "
		i+=msg.length();
		int k=partHeader[it].header.find(10,i);
		QCString tmp=partHeader[it].header.mid(i,k-i);
		int l=tmp.find(';');
		if(l!=-1){
			partHeader[it].contenttype=tmp.left(l);
			l=tmp.find("type=",l,TRUE);
			if(l==-1) partHeader[it].type="";
			else{
				partHeader[it].type=tmp.mid(l+6);
				partHeader[it].type=partHeader[it].type.left(partHeader[it].type.length()-1);
			}
		}else{
			partHeader[it].contenttype=tmp;
			partHeader[it].type="";
		}
	}
	//kdDebug(  ) << "[KmhtmlEntry::readPartHeader]: contenttype=" << partHeader[it].contenttype << endl;
	if(it==parts-1){
		// find start etc. for message header
		partHeader->start=0;
		partHeader->end=f.length();
		partHeader->itStart=partHeader->header.length();
		while(f.mid(partHeader->itStart,1)==QCString("-") || f.mid(partHeader->itStart,1)==controlChars[0] || f.mid(partHeader->itStart,1)==controlChars[1]) partHeader->itStart++;
		partHeader->itStart+=partHeader->boundary.length();
		while(f.mid(partHeader->itStart,1)==controlChars[0] || f.mid(partHeader->itStart,1)==controlChars[1]) partHeader->itStart++;
	}
	//kdDebug() << "[KmhtmlArchive::readPartHeader]: boundary=" << partHeader[it].boundary << endl;
	//kdDebug() << "[KmhtmlArchive::readPartHeader]: contenttype=" << partHeader[it].contenttype << endl;
	//kdDebug() << "[KmhtmlArchive::readPartHeader]: type=" << partHeader[it].type << endl;
	return TRUE;
}
/*================================
================================*/
bool KmhtmlArchive::findLineEnd(){
	if(f.find(13)!=-1) return TRUE;
	return FALSE;
}
/*================================
================================*/
QCString KmhtmlArchive::removeControlChars(QCString msg){
	int en;
	int st=0;
	while(msg.mid(st,1)==controlChars[0] || msg.mid(st,1)==controlChars[1]) st++;
	en=msg.length()-1;
	while(msg.mid(en,1)==controlChars[0] || msg.mid(en,1)==controlChars[1]) en--;
	return msg.mid(st,en-st+1);
}

/*================================
================================*/
void KmhtmlArchive::evalItemHeader(int id){
	QCString msg;
	int j;
	msg="Content-Transfer-Encoding:";  // --------------- read Content-Transfer-Encoding
	j=item[id].header.find(msg,0,TRUE);
	if(j==-1) item[id].encoding=-1;
	else{
		j++; // to include space ": "
		j+=msg.length();
		int k=item[id].header.find(10,j);
		QCString tmp=trim(removeControlChars(item[id].header.mid(j,k-j)));
		//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: encoding=\"" << tmp << "\""<< endl;
		if(tmp==QCString("quoted-printable")) item[id].encoding=0;
		else if(tmp==QCString("base64")) item[id].encoding=1;
		else if(tmp==QCString("uuencode")) item[id].encoding=2;
		else if(tmp==QCString("7bit") || f.mid(j,k-j-1)==QCString("7-bit")) item[id].encoding=3;
		else if(tmp==QCString("8bit") || f.mid(j,k-j-1)==QCString("8-bit")) item[id].encoding=4;
		else{
			item[id].encoding=-1;
			//kdDebug()<< "[KmhtmlArchive::findHeaderPart()]: unknown encoding " << endl;
		}
		//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: encoding=" << item[id].encoding << endl;
	}
	msg="Content-Type:"; // --------------- read Content-Type
	j=item[id].header.find(msg,0,TRUE);
	if(j==-1){
		item[id].mimeType="";
		//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: mimeType="<<endl;
	}else{
		j++; // to include space ": "
		j+=msg.length();
		int k=item[id].header.find(10,j);
		QCString tmp=trim(removeControlChars(item[id].header.mid(j,k-j)));
		int l=tmp.find(';');
		if(l!=-1){
			item[id].mimeType=tmp.left(l);
			l=tmp.find("charset=",l,TRUE);
			if(l==-1) item[id].charset="";
			else{
				item[id].charset=tmp.mid(l+8);
			}
		}else{
			item[id].mimeType=tmp;
			item[id].charset="";
		}
		//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: mimeType=" << item[id].mimeType << endl;
		//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: charset=" << item[id].charset << endl;
		//if(item[id].mimeType!=type) item[id].fixed=TRUE;
	}
	msg="Content-Location:"; // --------------- read Content-Location
	j=item[id].header.find(msg,0,TRUE);
	if(j==-1) item[id].contentLocation="";
	else{
		j++; // to include space ": "
		j+=msg.length();
		int k=item[id].header.find(10,j);
		item[id].contentLocation=trim(removeControlChars(item[id].header.mid(j,k-j)));
	}
	//item[i].itStart=j+item[i].contentLocation.length()+4;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: contentLocation=" << item[id].contentLocation << endl;
	msg="Content-ID:"; // --------------- read Content-ID
	j=item[id].header.find(msg,0,TRUE);
	if(j==-1){ item[id].contentID="";
	}else{
		j+=2; // to include ": <"
		j+=msg.length();
		int k=item[id].header.find('>',j);
		item[id].contentID=QString(item[id].header.mid(j,k-j));
	}
	//item[i].itStart=j+item[i].contentLocation.length()+4;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: Content-ID=" << item[id].contentID << endl;
	msg="filename="; // --------------- read filename (Content-Disposition)
	j=item[id].header.find(msg,0,TRUE);
	if(j==-1) item[id].dispFileName="";
	else{
		j++; // to include '"'
		j+=msg.length();
		int k=item[id].header.find('"',j);
		item[id].dispFileName=trim(removeControlChars(item[id].header.mid(j,k-j)));
	}
	//item[i].itStart=j+item[i].contentLocation.length()+4;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: dispFileName=" << item[id].dispFileName << endl;


	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< item[id].encoding << endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< item[id].mimeType << endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< item[id].contentLocation << endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< item[id].contentID << endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< item[id].dispFileName << endl;
	
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]"<< endl;
	
	// check header
	if(item[id].encoding==-1 && item[id].mimeType.isEmpty() && item[id].contentLocation.isEmpty() && item[id].contentID.isEmpty() && item[id].dispFileName.isEmpty()){
		// this is the alternative text for brousers without mhtml support
		item[id].header="";
		item[id].itStart=item[id].start;
		item[id].encoding=3;
		item[id].mimeType="text/plain";
		item[id].name="alternative-text.txt";
		item[id].alias="alternative-text.txt";
		item[id].hasExtension=TRUE;
	}else if(item[id].encoding==3 && item[id].mimeType=="text/plain" && item[id].contentLocation.isEmpty() && item[id].contentID.isEmpty() && item[id].dispFileName.isEmpty()){
		item[id].name="alternative-text.txt";
		item[id].alias="alternative-text.txt";
		item[id].hasExtension=TRUE;
	}else{
		// extract file name
		item[id].hasExtension=FALSE;
		if(!item[id].dispFileName.isEmpty()){
			item[id].name=item[id].dispFileName;
			//item[id].alias=item[id].dispFileName;
			item[id].hasExtension=TRUE;
		}else{
			item[id].alias="";
			if(!item[id].contentLocation.isEmpty()){
				int c=item[id].contentLocation.findRev("/");
				if(c!=-1) item[id].name=item[id].contentLocation.right(item[id].contentLocation.length()-c-1);
				else item[id].name=item[id].contentLocation;
				c=item[id].contentLocation.findRev(QChar(0x5c));
				if(c!=-1) item[id].name=item[id].contentLocation.right(item[id].contentLocation.length()-c-1);
				c=-1;
				while ((c=item[id].name.find("?",c+1))!=-1) item[id].name=item[id].name.right(item[id].name.length()-c-1);
				c=-1;
				while ((c=item[id].name.find("=",c+1))!=-1) item[id].name=item[id].name.right(item[id].name.length()-c-1);
				//kdDebug()<< "[KmhtmlArchive::readItems()]: item[id].name = " << item[id].name << endl;
				c=item[id].name.find(".");
				if(c!=-1) item[id].hasExtension=TRUE;
			}else if(!item[id].contentID.isEmpty()){
				int papaki=item[id].contentID.find("@");
				int dot=item[id].contentID.find(".");
				if(dot<papaki && dot!=-1) item[id].fixedName=item[id].contentID.left(dot);
			}
			// Calculate item size
			item[id].encodedSize=item[id].end-item[id].itStart;
			if(item[id].name.right(3).lower()==QString(".js")) item[id].mimeType="text/plain";
		}
	}
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: name=" << item[id].name << endl;
	//kdDebug() << "[KmhtmlArchive::findHeaderPart()]: alias=" << item[id].alias << endl;
}
/*================================
================================*/
QCString KmhtmlArchive::trim(QCString tmp){
	int st=0;
	int en=0;
	for(uint i=0;i<tmp.length();i++){
		if(tmp.mid(i,1)==" " || tmp.mid(i,1)=="\t") st++;
		else break;
	}
	for(uint i=tmp.length()-1;i<=0;i--){
		if(tmp.mid(i,1)==" " || tmp.mid(i,1)=="\t") en++;
		else break;
	}
	return tmp.mid(st,tmp.length()-st-en);
}
// KmhtmlEntry
/*================================
================================*/
// KmhtmlEntry::KmhtmlEntry(){
// 	archive=0;
// 	it=-1;
// 	//onlyfname="";
// 	m_exists=FALSE;
// }
/*================================
================================*/
KmhtmlEntry::KmhtmlEntry(KmhtmlArchive* arch,QString fileName){
	if(arch){
		archive=arch;
		fname=fileName;
		//kdDebug() << "[KmhtmlEntry::KmhtmlEntry()]: findItemByName" << endl;
		if(!findItemByName()){
			kdDebug() << "[KmhtmlEntry::KmhtmlEntry()]: does not exist!!! " << fname << endl;
			m_exists=FALSE;
			return;
		}
		else m_exists=TRUE;
		kdDebug() << "[KmhtmlEntry::KmhtmlEntry()]: exists!!!" << endl;
	}else{
		archive=0;
		it=-1;
		fname="";
		m_exists=FALSE;
	}
}
/*================================
================================*/
KmhtmlEntry::KmhtmlEntry(KmhtmlArchive* arch, int item){
	if(arch){
		archive=arch;
		it=item;
		if(!findItemById()){
			m_exists=FALSE;
			return;
		}
		else m_exists=TRUE;
		//kdDebug()<< "KmhtmlEntry::KmhtmlEntry(KmhtmlArchive* arch, int item): it = " << it << endl;
	}else{
		archive=0;
		it=-1;
		fname="";
	}
}
/*================================
================================*/
KmhtmlEntry::~KmhtmlEntry(){
}
/*================================
================================*/
QDateTime KmhtmlEntry::datetime() const{
	QDateTime d;
	d.setTime_t( archive->m_mtime );
	return d;
}
/*================================
================================*/
bool KmhtmlEntry::findItemById(){
	fname="";
	if(it>=0 && it<archive->items) fname=archive->item[it].name;
	else return FALSE;
	return TRUE;
}
/*================================
================================*/
bool KmhtmlEntry::findItemByName(){
	it=-1;
	if(fname.isEmpty()) return FALSE;
	else{
		for(int i=0;i<archive->items;i++) if(fname==archive->item[i].name || fname==archive->item[i].alias) it=i;
		if(it==-1) return FALSE;
	}
	return TRUE;
}
// /*================================
// ================================*/
// void KmhtmlEntry::fixName(){
// 	int c=-1;
// 	c=fname.findRev("/");
// 	if(c==-1) onlyfname=fname;
// 	else onlyfname=fname.mid(c+1);
// }
// /*================================
// ================================*/
// QString KmhtmlEntry::name(){
// 	int c=falias.findRev("/");
// 	//kdDebug()<<"[KmhtmlEntry::name()]: falias=" << falias << endl;
// 	//kdDebug()<<"[KmhtmlEntry::name()]: item[i].alias=" << archive->item[it].alias << endl;
// 	if(c==-1) return falias;
// 	else return falias.mid(c+1);
// }
/*================================
================================*/
QString KmhtmlEntry::escapeString(QString str){
	QString ret=QString::null;
	uint strLength=str.length();
	for(int i=0;(uint)i<strLength;i++){
		QChar c=str.at((uint)i);
		if(c==QChar('\\') || c==QChar('/') || c==QChar('(') || c==QChar(')') || 
			c==QChar('[') || c==QChar(']') || c==QChar('{') || c==QChar('}') || 
			c==QChar('$') || c==QChar('^'))
			ret.append(QChar('\\'));
		ret.append(str.mid(i,1));
	}
	return ret;
}
/*================================
================================*/
QByteArray* KmhtmlEntry::data(){
	//kdDebug()<< "[KmhtmlEntry::data()]: it=" << it << endl;
	if(archive->item[it].mimeType==archive->type && !archive->item[it].fixed){
		//kdDebug()<< "[KmhtmlEntry::data()]: I'm in" << endl;
		QCString str;
		str=archive->item[it].data;
		QRegExp findReg;
		findReg.setCaseSensitive(TRUE);
		findReg.setMinimal(TRUE);
		findReg.setWildcard(FALSE);
		for(int j=0;j<archive->count();j++){
		//for(int j=archive->count()-1;j>=0;j--){
			KmhtmlEntry arcF(archive, j);
			if(arcF.mimeType()!=archive->Type()){
				// fix links
				if(!arcF.contentLocation().isEmpty()){
					QString loc=escapeString(arcF.contentLocation());
					findReg.setPattern(loc);
					str.replace(findReg,arcF.name());
				}
				QString arcFname=escapeString(arcF.name());
				//kdDebug()<< "arcFname=" << arcFname << endl;
				if(!arcF.originalName().isEmpty()){
					QString arcForiginalName=escapeString(arcF.originalName());
					//kdDebug()<< "arcForiginalName=" << arcForiginalName << endl;
					if(arcF.originalName()!=arcF.name()){
						//kdDebug()<<"replacing "<<arcF.originalName() << " with " << arcF.name() <<endl;
						
						
						//findReg.setPattern(arcForiginalName);
						int found=0;
						while(found!=-1){
							found=str.find(arcF.originalName().utf8(),found+1);
							if(found!=-1){
								int endFound=found;
								endFound+=arcF.originalName().length();
								//kdDebug()<<str.mid(found,endFound-found)<<endl;
								//kdDebug()<<str.mid(endFound,1)<<endl;
								if(str.mid(endFound,1)==QCString(">")
									|| str.mid(endFound,1)==QCString(" ")
									|| str.mid(endFound,1)==QCString("\'")
									|| str.mid(endFound,1)==QCString("\"")
									|| str.mid(endFound,1)==QCString("\r")
									|| str.mid(endFound,1)==QCString("\n")
									|| str.mid(endFound,1)==QCString("\t")) {
									str.replace(found,endFound-found,arcF.name().utf8());
									//kdDebug()<<"replacing "<<arcF.originalName() << " with " << arcF.name() <<endl;
								}
							}
						}
						//str.replace(findReg,arcF.name());
					}
					findReg.setPattern(QString("http[^ ]*") + arcFname );
					str.replace(findReg,arcF.name());
					findReg.setPattern(QString("file://.*") + arcFname );
					str.replace(findReg,arcF.name());
					findReg.setPattern(QString("ftp://.*") + arcFname );
					str.replace(findReg,arcF.name());
				}
				if(!archive->item[j].contentID.isEmpty()){
					QString contID=escapeString(archive->item[j].contentID);
					contID.prepend("cid:");
					findReg.setPattern( contID );
					str.replace(findReg,arcF.name());
				}
			}
		}
		//kdDebug(  ) << "[mhtmlProtocol::data]: done!!!" << endl;
		QByteArray result;
		result=str;
		result.resize(result.size()-1);
		archive->item[it].decodedSize=result.size();
		archive->item[it].data=result;
		archive->item[it].fixed=TRUE;
	}
	//kdDebug()<< "[KmhtmlEntry::data()]: return normal" << endl;
	return &archive->item[it].data;
}
#include <kmhtmlarchive.moc>
